import unittest
from froc_analysis.utils import get_iou_score


class TestIoU(unittest.TestCase):
    def setUp(self):
        self.gt_box = [100, 100, 100, 100]  # x, y, w, h
        self.pr_box1 = [100, 100, 50, 50]
        self.pr_box2 = [150, 150, 25, 25]
        self.pr_box3 = [200, 200, 100, 100]

    def test_calculate_overlapping_box_from_same_xy(self):
        iou = get_iou_score(self.gt_box, self.pr_box1)
        self.assertEqual(iou, 1 / 4)

    def test_calculate_overlapping_box_from_center(self):
        iou = get_iou_score(self.gt_box, self.pr_box2)
        self.assertEqual(iou, 1 / 16)

    def test_calculate_non_overlapping_boxes(self):
        iou = get_iou_score(self.gt_box, self.pr_box3)
        self.assertEqual(iou, 0.0)



import unittest
import numpy as np

from froc_analysis.utils import update_scores


class TestUpdateScore(unittest.TestCase):
    def setUp(self):
        self.json_data = [{"score": float(score) / 100.0} for score in range(0, 105, 5)]

    def test_thresholding(self):
        preds = update_scores(self.json_data, 0.5)
        self.assertEqual(len(preds), 10)
        preds = update_scores(self.json_data, 1.0)
        self.assertEqual(len(preds), 0)



import json
from pathlib import Path


def load_json_from_file(file_path):
    with Path(file_path).open() as fp:
        data = json.load(fp)
    return data


def update_scores(json_data: list, score_thres: float) -> list:
    preds = []
    for ind, pred in enumerate(json_data):
        if pred["score"] > score_thres:
            preds.append(pred)
    return preds


def get_iou_score(gt_box: list, pr_box: list) -> float:
    """IoU score between GT and prediction boxes.

    Arguments:
        gt_box {list} -- [x, y, w, h] of ground-truth lesion
        pr_box {list} -- [x, y, w, h] of prediction bounding box

    Returns:
        score {float} -- intersection over union score of the two boxes
    """
    gt_x, gt_y, gt_w, gt_h = gt_box
    pr_x, pr_y, pr_w, pr_h = pr_box

    xA = max(gt_x, pr_x)
    xB = min(gt_x + gt_w, pr_x + pr_w)
    yA = max(gt_y, pr_y)
    yB = min(gt_y + gt_h, pr_y + pr_h)

    intersection = max((xB - xA), 0) * max((yB - yA), 0)
    if intersection == 0:
        return 0.0

    gt_area = gt_w * gt_h
    pr_area = pr_w * pr_h

    return intersection / (gt_area + pr_area - intersection)


def build_pr_id2annotations(pr: list) -> dict:
    """Build image to annotation dictionary based on list of predictions.

    Arguments:
        pr {list} -- List of predictions from Detectron2 (coco-instance-results.json)

    Returns:
        id_to_annotation {dict} -- Image IDs to predicted annotations.
    """
    id_to_annotation = dict()
    for annotation in pr:
        id_to_annotation[annotation["image_id"]] = []
    for annotation in pr:
        id_to_annotation[annotation["image_id"]].append(annotation)
    return id_to_annotation


def build_gt_id2annotations(gt: dict) -> dict:
    """Build image to annotation dictionary based on the ground truth dataset.

    Arguments:
        gt {dict} -- COCO dataset of ground truth annotations.

    Returns:
        id_to_annotation {dict} -- Image IDs to ground-truth annotations.
    """
    id_to_annotation = dict()
    for image in gt["images"]:
        id_to_annotation[image["id"]] = []
    for annotation in gt["annotations"]:
        id_to_annotation[annotation["image_id"]].append(annotation)
    return id_to_annotation



import unittest
import os
import mock
import warnings

# remove tqdm logging while testing
def notqdm(iterable, *args, **kwargs):
    return iterable


from froc_analysis.froc_curve import froc_point
from froc_analysis import generate_froc_curve


class TestFrocCurve(unittest.TestCase):
    def setUp(self):
        self.gt_ann_path = f"{os.getcwd()}/tests/example_coco_data/face-mask.json"
        self.pr_prefect = (
            f"{os.getcwd()}/tests/example_coco_data/perfect_predictions.json"
        )
        self.pr_single_error_per_image_high_conf = f"{os.getcwd()}/tests/example_coco_data/single_error_per_image_predictions.json"
        self.pr_single_error_per_image_low_conf = f"{os.getcwd()}/tests/example_coco_data/single_error_per_image_low_confidence_predictions.json"
        warnings.simplefilter("ignore", category=UserWarning)

    def test_perfect_predictions_for_all_score_thresholds(self):
        for score_thresh in range(0, 100, 5):
            score_thresh = float(score_thresh) / 100.0
            stats = froc_point(
                self.gt_ann_path,
                self.pr_prefect,
                score_thresh,
                use_iou=False,
                iou_thres=0.0,
            )
            for cat_id in stats:
                self.assertEqual(stats[cat_id]["LL"] / stats[cat_id]["n_lesions"], 1.0)
                self.assertEqual(stats[cat_id]["NL"] / stats[cat_id]["n_images"], 0.0)

    def test_single_error_per_image_predictions_for_all_score_thresholds(self):
        for score_thresh in range(0, 100, 5):
            score_thresh = float(score_thresh) / 100.0
            stats = froc_point(
                self.gt_ann_path,
                self.pr_single_error_per_image_high_conf,
                score_thresh,
                use_iou=False,
                iou_thres=0.0,
            )
            for cat_id in stats:
                self.assertEqual(
                    stats[cat_id]["LL"] / stats[cat_id]["n_lesions"],
                    1.0 - stats[cat_id]["n_images"] / stats[cat_id]["n_lesions"],
                )

    def test_single_error_per_image_low_confidence_predictions_for_all_score_thresholds(
        self
    ):
        for score_thresh in range(0, 100, 5):
            score_thresh = float(score_thresh) / 100.0
            stats = froc_point(
                self.gt_ann_path,
                self.pr_single_error_per_image_low_conf,
                score_thresh,
                use_iou=False,
                iou_thres=0.0,
            )
            for cat_id in stats:
                specificity = (
                    1.0 - stats[cat_id]["n_images"] / stats[cat_id]["n_lesions"]
                )
                if score_thresh >= 0.25:
                    specificity = 0.0
                self.assertEqual(
                    stats[cat_id]["LL"] / stats[cat_id]["n_lesions"], specificity
                )

    @mock.patch("froc_analysis.froc_curve.tqdm", notqdm)
    def test_froc_curve_generation_for_perfect_score(self):
        lls, nlls = generate_froc_curve(
            self.gt_ann_path,
            self.pr_prefect,
            use_iou=False,
            iou_thres=0.0,
            n_sample_points=10,
            plot_title=None,
            plot_output_path=None,
        )
        self.assertListEqual(lls[8], [1.0 for _ in range(10)])
        self.assertListEqual(nlls[8], [0.0 for _ in range(10)])

    @mock.patch("froc_analysis.froc_curve.tqdm", notqdm)
    def test_froc_curve_generation_for_single_error_per_image(self):
        lls, nlls = generate_froc_curve(
            self.gt_ann_path,
            self.pr_single_error_per_image_high_conf,
            use_iou=False,
            iou_thres=0.0,
            n_sample_points=10,
            plot_title=None,
            plot_output_path=None,
        )
        self.assertListEqual(lls[8], [0.7 for _ in range(10)])
        self.assertListEqual(nlls[8], [0.0 for _ in range(10)])

    @mock.patch("froc_analysis.froc_curve.tqdm", notqdm)
    def test_froc_curve_generation_for_single_error_per_image_low_confidence(self):
        lls, nlls = generate_froc_curve(
            self.gt_ann_path,
            self.pr_single_error_per_image_low_conf,
            use_iou=False,
            iou_thres=0.0,
            n_sample_points=10,
            plot_title=None,
            plot_output_path=None,
        )
        self.assertListEqual(lls[8], [0.7 for _ in range(3)] + [0.0 for _ in range(7)])
        self.assertListEqual(nlls[8], [0.0 for _ in range(10)])

    @mock.patch("froc_analysis.froc_curve.tqdm", notqdm)
    def test_froc_curve_saving_to_plot(self):
        generate_froc_curve(
            self.gt_ann_path,
            self.pr_prefect,
            use_iou=False,
            iou_thres=0.0,
            n_sample_points=10,
            plot_title="perfect scores",
            plot_output_path="/tmp/froc_perfect_scores.png",
        )



from collections import Counter
from .utils import get_iou_score

def init_stats(gt: dict, categories: dict) -> dict:
    """Initializing the statistics before counting leasion
       and non-leasion localiazations.

    Arguments:
        gt {dict} -- Ground truth COCO dataset
        categories {dict} -- Dictionary of categories present in the COCO dataset

    Returns:
        stats {dict} -- Statistics to be updated, containing every information
                        necessary to evaluate a single FROC point
    """
    stats = {
        cat["id"]: {
            "name": cat["name"],
            "LL": 0,
            "NL": 0,
            "n_images": [],
            "n_lesions": 0,
        }
        for cat in categories
    }

    for annotation in gt["annotations"]:
        category_id = annotation["category_id"]
        stats[category_id]["n_lesions"] += 1

    for image in gt["images"]:
        image_id = image["id"]
        for cat_id in stats:
            stats[cat_id]["n_images"].append(image_id)

    for cat_id in stats:
        stats[cat_id]["n_images"] = len(stats[cat_id]["n_images"])

    return stats


def update_stats(
    stats: dict,
    gt_id_to_annotation: dict,
    pr_id_to_annotation: dict,
    categories: dict,
    use_iou: bool,
    iou_thres: float,
):
    """Updating statistics as going through images of the dataset.

    Arguments:
        stats {dict} -- FROC statistics
        gt_id_to_annotation {dict} -- Ground-truth image IDs to annotations.
        pr_id_to_annotation {dict} -- Prediction image IDs to annotations.
        categories {dict} -- COCO categories dictionary.
        use_iou {bool} -- Whether or not to use iou thresholding.
        iou_thres {float} -- IoU threshold when using IoU thresholding.

    Returns:
        stats {dict} -- Updated FROC statistics
    """
    for image_id in gt_id_to_annotation:
        n_is_ll = dict()
        for cat in categories:
            n_is_ll[cat["id"]] = 0

        for gt_ann in gt_id_to_annotation[image_id]:
            for pred_ann in pr_id_to_annotation.get(image_id, []):
                if gt_ann["category_id"] != pred_ann["category_id"]:
                    continue

                if use_iou:
                    iou_score = get_iou_score(gt_ann["bbox"], pred_ann["bbox"])
                    if iou_thres < iou_score:
                        stats[gt_ann["category_id"]]["LL"] += 1
                        n_is_ll[gt_ann["category_id"]] += 1
                        break
                else:
                    gt_x, gt_y, gt_w, gt_h = gt_ann["bbox"]

                    pr_x, pr_y, pr_w, pr_h = pred_ann["bbox"]
                    pr_bbox_center = pr_x + pr_w / 2, pr_y + pr_h / 2

                    if (
                        pr_bbox_center[0] >= gt_x
                        and pr_bbox_center[0] <= gt_x + gt_w
                        and pr_bbox_center[1] >= gt_y
                        and pr_bbox_center[1] <= gt_y + gt_h
                    ):
                        stats[gt_ann["category_id"]]["LL"] += 1
                        n_is_ll[gt_ann["category_id"]] += 1
                        break

        cat_to_n_pred = Counter(
            [pr_ann["category_id"] for pr_ann in pr_id_to_annotation.get(image_id, [])]
        )

        difference = cat_to_n_pred - Counter(n_is_ll)
        for cat_id in difference:
            stats[cat_id]["NL"] += difference[cat_id]

    return stats



from .froc_stats import init_stats, update_stats
from .utils import (
    load_json_from_file,
    update_scores,
    build_gt_id2annotations,
    build_pr_id2annotations,
)
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm


def froc_point(gt_ann, pr_ann, score_thres, use_iou, iou_thres):
    gt = load_json_from_file(gt_ann)
    pr = load_json_from_file(pr_ann)

    pr = update_scores(pr, score_thres)

    categories = gt["categories"]

    stats = init_stats(gt, categories)

    gt_id_to_annotation = build_gt_id2annotations(gt)
    pr_id_to_annotation = build_pr_id2annotations(pr)

    stats = update_stats(
        stats, gt_id_to_annotation, pr_id_to_annotation, categories, use_iou, iou_thres
    )

    return stats


def generate_froc_curve(
    gt_ann, pr_ann, use_iou=False,
    iou_thres=.5, n_sample_points=50,
    plot_title='FROC curve', plot_output_path='froc.png'
):
    lls_accuracy = {}
    nlls_per_image = {}

    for score_thres in tqdm(np.linspace(0.0, 1.0, n_sample_points, endpoint=False)):
        stats = froc_point(gt_ann, pr_ann, score_thres, use_iou, iou_thres)
        for category_id in stats:
            if lls_accuracy.get(category_id, None):
                lls_accuracy[category_id].append(
                    stats[category_id]["LL"] / stats[category_id]["n_lesions"]
                )
            else:
                lls_accuracy[category_id] = []
                lls_accuracy[category_id].append(
                    stats[category_id]["LL"] / stats[category_id]["n_lesions"]
                )

            if nlls_per_image.get(category_id, None):
                nlls_per_image[category_id].append(
                    stats[category_id]["NL"] / stats[category_id]["n_images"]
                )
            else:
                nlls_per_image[category_id] = []
                nlls_per_image[category_id].append(
                    stats[category_id]["NL"] / stats[category_id]["n_images"]
                )

    if plot_title:
        plt.figure(figsize=(12, 12))

    for category_id in lls_accuracy:
        lls = lls_accuracy[category_id]
        nlls = nlls_per_image[category_id]
        if plot_title:
            plt.semilogx(nlls, lls, "x--", label=stats[category_id]["name"])

    if plot_title:
        plt.legend(loc="lower right")

        plt.title(plot_title)
        plt.ylabel("Sensitivity")
        plt.xlabel("FP / per image")

        plt.tight_layout()

        plt.savefig(plot_output_path, dpi=50)
    else:
        return lls_accuracy, nlls_per_image



from copy import deepcopy
import json
import matplotlib.pyplot as plt
from tqdm import tqdm
import random
import os
import numpy as np

from .froc_curve import generate_froc_curve


def generate_bootstrap_curves(
    gt_ann,
    pr_ann,
    n_bootstrap_samples=5,
    use_iou=False,
    iou_thres=.5,
    n_sample_points=50,
    plot_title='Bootstrap FROC',
    plot_output_path='froc_bootstrapped.png',
):
    with open(gt_ann, "r+") as fp:
        GT_ANN = json.load(fp)

    with open(pr_ann, "r+") as fp:
        PRED_ANN = json.load(fp)

    n_images = len(GT_ANN["images"])

    plt.figure(figsize=(15, 15))

    collected_frocs = {"lls": {}, "nlls": {}}

    for _ in tqdm(range(n_bootstrap_samples)):
        selected_images = random.choices(GT_ANN["images"], k=n_images)
        bootstrap_gt = deepcopy(GT_ANN)

        del bootstrap_gt["images"]

        bootstrap_gt["images"] = selected_images

        gt_annotations = bootstrap_gt["annotations"]

        del bootstrap_gt["annotations"]

        bootstrap_gt["annotations"] = []
        for _gt_ann_ in gt_annotations:
            img_id = _gt_ann_["image_id"]
            for selected_image in selected_images:
                if selected_image["id"] == img_id:
                    bootstrap_gt["annotations"].append(_gt_ann_)

        with open("/tmp/tmp_bootstrap_gt.json", "w") as fp:
            json.dump(bootstrap_gt, fp)

        with open("/tmp/tmp_bootstrap_pred.json", "w") as fp:
            json.dump(PRED_ANN, fp)

        tmp_gt_ann = "/tmp/tmp_bootstrap_gt.json"
        tmp_pred_ann = "/tmp/tmp_bootstrap_pred.json"

        lls, nlls = generate_froc_curve(
            tmp_gt_ann,
            tmp_pred_ann,
            use_iou,
            iou_thres,
            n_sample_points,
            plot_title=None,
            plot_output_path=None,
        )

        for cat_id in lls:
            plt.semilogx(lls[cat_id], nlls[cat_id], "-", alpha=0.05)
            collected_frocs["lls"].get(cat_id, []).append(lls[cat_id])
            collected_frocs["nlls"].get(cat_id, []).append(nlls[cat_id])

        collected_frocs.append((lls, nlls))

    mean_froc_lls = {}
    mean_froc_nlls = {}

    for cat_id in collected_frocs["lls"]:
        mean_froc_lls[cat_id] = np.mean(
            np.array(collected_frocs["lls"][cat_id]).reshape(
                n_bootstrap_samples, n_sample_points
            ),
            axis=0,
        )
        mean_froc_nlls[cat_id] = np.mean(
            np.array(collected_frocs["nlls"][cat_id]).reshape(
                n_bootstrap_samples, n_sample_points
            ),
            axis=0,
        )

    mean_froc_curve = {}
    for cat_id in collected_frocs["lls"]:
        mean_froc_curve[cat_id] = np.stack(
            (mean_froc_nlls[cat_id], mean_froc_lls[cat_id]), axis=-1
        )

        plt.semilogx(mean_froc_curve[cat_id][:, 0], mean_froc_curve[cat_id][:, 1], "x-")

    plt.xlabel("FP/image")
    plt.ylabel("Sensitivity")

    os.remove("/tmp/tmp_bootstrap_gt.json")
    os.remove("/tmp/tmp_bootstrap_pred.json")

    plt.title(plot_title)

    plt.savefig(plot_output_path, dpi=100)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="coco_froc_analysis",
    version="0.0.23",
    author="Alex Olar",
    author_email="olaralex666@gmail.com",
    description="A small package that evaluates COCO detection results from OpenMMLab and Detectron(2).",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/qbeer/coco-froc-analysis",
    project_urls={
        "Bug Tracker": "https://github.com/qbeer/coco-froc-analysis/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
)


import unittest

from froc_analysis.utils import build_pr_id2annotations, build_gt_id2annotations


class TestId2Annotation(unittest.TestCase):
    def setUp(self):
        self.gt = {
            "images": [{"id": 1}, {"id": 2}],
            "annotations": [
                {"image_id": 1},
                {"image_id": 1},
                {"image_id": 1},
                {"image_id": 2},
            ],
        }
        self.pr = [
            {"image_id": 1},
            {"image_id": 1},
            {"image_id": 1},
            {"image_id": 1},
            {"image_id": 1},
            {"image_id": 2},
            {"image_id": 2},
            {"image_id": 2},
        ]

    def test_build_ground_truth_id2annotations_dict(self):
        id2annotation = build_gt_id2annotations(self.gt)
        self.assertEqual(len(id2annotation[1]), 3)
        self.assertEqual(len(id2annotation[2]), 1)

    def test_build_prediction_id2annotations_dict(self):
        id2annotation = build_pr_id2annotations(self.pr)
        self.assertEqual(len(id2annotation[1]), 5)
        self.assertEqual(len(id2annotation[2]), 3)
